/*
 * Copyright (C) 2003-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file mp3_afh.c para_server's mp3 audio format handler */

/*
 * This file is based in part on mp3tech.c and mp3tech.h, Copyright (C)
 * 2000-2001 Cedric Tefft <cedric@earthling.net>, which in turn is based
 * in part on
 *
 *	* MP3Info 0.5 by Ricardo Cerqueira <rmc@rccn.net>
 *	* MP3Stat 0.9 by Ed Sweetman <safemode@voicenet.com> and
 *			 Johannes Overmann <overmann@iname.com>
 */

#include "server.h"
#include "error.h"
#include "string.h"

/** \cond some defines and structs which are only used in this file */

/*
 * MIN_CONSEC_GOOD_FRAMES defines how many consecutive valid MP3 frames we need
 * to see before we decide we are looking at a real MP3 file
 */
#define MIN_CONSEC_GOOD_FRAMES 4
#define FRAME_HEADER_SIZE 4
#define MIN_FRAME_SIZE 21

struct mp3header {
	unsigned long sync;
	unsigned int version;
	unsigned int layer;
	unsigned int crc;
	unsigned int bitrate;
	unsigned int freq;
	unsigned int padding;
	unsigned int mode;
	unsigned int copyright;
	unsigned int original;
	unsigned int emphasis;
};

struct id3tag {
	char title[31];
	char artist[31];
	char album[31];
	char year[5];
	char comment[31];
};

struct mp3info {
	struct mp3header header;
	int id3_isvalid;
	struct id3tag id3;
	int vbr;
};

/** \endcond */
static const int frequencies[3][4] = {
	{22050,24000,16000,50000}, /* MPEG 2.0 */
	{44100,48000,32000,50000}, /* MPEG 1.0 */
	{11025,12000,8000,50000} /* MPEG 2.5 */
};

static const int mp3info_bitrate[2][3][14] = {
{ /* MPEG 2.0 */
	{32,48,56,64,80,96,112,128,144,160,176,192,224,256}, /* layer 1 */
	{8,16,24,32,40,48,56,64,80,96,112,128,144,160}, /* layer 2 */
	{8,16,24,32,40,48,56,64,80,96,112,128,144,160} /* layer 3 */
},

{ /* MPEG 1.0 */
	{32,64,96,128,160,192,224,256,288,320,352,384,416,448}, /* layer 1 */
	{32,48,56,64,80,96,112,128,160,192,224,256,320,384}, /* layer 2 */
	{32,40,48,56,64,80,96,112,128,160,192,224,256,320} /* layer 3 */
}
};

static const int frame_size_index[] = {24000, 72000, 72000};
static const char *mode_text[] = {"stereo", "joint stereo", "dual channel", "mono", "invalid"};

static struct mp3info mp3;

static int header_frequency(struct mp3header *h)
{
	if (h->version > 2 || h->freq > 3)
		return -E_HEADER_FREQ;
	return frequencies[h->version][h->freq];
}

static const char *header_mode(struct mp3header *h)
{
	if (h->mode > 4)
		h->mode = 4; /* invalid */
	return mode_text[h->mode];
}

static int header_channels(struct mp3header *h)
{
	if (h->mode > 3)
		return 0;
	if (h->mode < 3)
		return 2;
	return 1;
}

static int header_bitrate(struct mp3header *h)
{
	if (!h->layer || h->layer > 3 || h->bitrate > 14 || !h->bitrate)
		return -E_HEADER_BITRATE;
	return mp3info_bitrate[h->version & 1][3 - h->layer][h->bitrate - 1];
}

static int frame_length(struct mp3header *header)
{
	int hb, hf = header_frequency(header);

	if (hf < 0)
		return hf;
	hb = header_bitrate(header);
	if (hb < 0)
		return hb;
	if (header->sync != 0xFFE || header->layer > 3)
		return -E_FRAME;
	return frame_size_index[3 - header->layer] *
		((header->version & 1) + 1) * hb / hf
		+ header->padding;
}

static void write_info_str(struct audio_format_info *afi)
{
	int v = mp3.id3_isvalid;

	snprintf(afi->info_string, MMD_INFO_SIZE,
		"audio_file_info1:%lu x %lums, %u kbit/s (%cbr) %i KHz %s\n"
		"audio_file_info2:%s, by %s\n"
		"audio_file_info3:A: %s, Y: %s, C: %s\n",
		afi->chunks_total,
		tv2ms(&afi->chunk_tv),
		afi->bitrate,
		mp3.vbr? 'v' : 'c',
		afi->frequency / 1000,
		header_mode(&mp3.header),
		v && *mp3.id3.title? mp3.id3.title : "(title tag not set)",
		v && *mp3.id3.artist? mp3.id3.artist : "(artist tag not set)",
		v && *mp3.id3.album? mp3.id3.album : "(album tag not set)",
		v && *mp3.id3.year? mp3.id3.year : "????",
		v && *mp3.id3.comment? mp3.id3.comment : "(comment tag not set)"
	);
}

/*
 * Remove trailing whitespace from the end of a string
 */
static char *unpad(char *string)
{
	char *pos = string + strlen(string) - 1;
	while (isspace(pos[0]))
		(pos--)[0] = 0;
	return string;
}

static int compare_headers(struct mp3header *h1,struct mp3header *h2)
{
	if ((*(uint*)h1) == (*(uint*)h2))
		return 1;
	if ((h1->version == h2->version) &&
			(h1->layer == h2->layer) &&
			(h1->crc == h2->crc) &&
			(h1->freq == h2->freq) &&
			(h1->mode == h2->mode) &&
			(h1->copyright == h2->copyright) &&
			(h1->original == h2->original) &&
			(h1->emphasis == h2->emphasis))
		return 1;
	return 0;
}

/*
 * get next MP3 frame header.
 *
 * On success, the header frame length is returned and the given header
 * structure that is filled in.  A return value of zero means that we did not
 * retrieve a valid frame header, and a negative return value indicates an
 * error.
 */
static int get_header(unsigned char *map, off_t numbytes, off_t *fpos,
	struct mp3header *header)
{
	int fl, ret;

	if (*fpos + FRAME_HEADER_SIZE > numbytes) {
		*fpos = numbytes - 1;
		header->sync = 0;
		return 0;
	}
	header->layer = (map[*fpos + 1] >> 1) & 3;
	header->sync = (((int)map[*fpos]<<4) | ((int)(map[*fpos + 1]&0xE0)>>4));
	if (map[*fpos + 1] & 0x10)
		header->version = (map[*fpos + 1] >> 3) & 1;
	else
		header->version = 2;
	if ((header->sync != 0xFFE) || (header->layer != 1)) {
		ret = 0;
		header->sync = 0;
		goto out;
	}
	header->crc = map[*fpos + 1] & 1;
	header->bitrate = (map[*fpos + 2] >> 4) & 0x0F;
	header->freq = (map[*fpos + 2] >> 2) & 0x3;
	header->padding = (map[*fpos + 2] >>1) & 0x1;
	header->mode = (map[*fpos + 3] >> 6) & 0x3;
	fl = frame_length(header);
	ret = (fl >= MIN_FRAME_SIZE)? fl : -E_FRAME_LENGTH;
out:
	*fpos += FRAME_HEADER_SIZE;
	return ret;
}

/*
 * find the next mp3 header
 *
 * Return the length of the next frame header or zero if the end of the file is
 * reached.
 */
static int mp3_seek_next_header(unsigned char *map, off_t numbytes, off_t *fpos)
{
	int k, l = 0, first_len;
	struct mp3header h, h2;
	long valid_start = 0;

	for (; *fpos < numbytes; (*fpos)++) {
		if (map[*fpos] != 0xff)
			continue;
		valid_start = *fpos;
		first_len = get_header(map, numbytes, fpos, &h);
		if (first_len <= 0)
			continue;
		*fpos += first_len - FRAME_HEADER_SIZE;
		for (k = 1; k < MIN_CONSEC_GOOD_FRAMES; k++) {
			if ((l = get_header(map, numbytes, fpos, &h2)) <= 0)
				break;
			if (!compare_headers(&h, &h2))
				break;
			*fpos += l - FRAME_HEADER_SIZE;
		}
		if (k == MIN_CONSEC_GOOD_FRAMES) {
			*fpos = valid_start;
			memcpy(&(mp3.header), &h2, sizeof(struct mp3header));
			return first_len;
		}
	}
	return 0;
}

static void mp3_get_id3(unsigned char *map, off_t numbytes, off_t *fpos)
{
	mp3.id3_isvalid = 0;
	mp3.id3.title[0] = '\0';
	mp3.id3.artist[0] = '\0';
	mp3.id3.album[0] = '\0';
	mp3.id3.comment[0] = '\0';
	mp3.id3.year[0] = '\0';
	if (numbytes < 128)
		return;
	*fpos = numbytes - 128;
	if (strncmp("TAG", (char *) map + *fpos, 3)) {
		PARA_INFO_LOG("%s", "no id3 tag\n");
		return;
	}
	*fpos = numbytes - 125;
	memcpy(mp3.id3.title, map + *fpos, 30);
	*fpos += 30;
	mp3.id3.title[30] = '\0';
	memcpy(mp3.id3.artist, map + *fpos, 30);
	*fpos += 30;
	mp3.id3.artist[30] = '\0';
	memcpy(mp3.id3.album, map + *fpos, 30);
	*fpos += 30;
	mp3.id3.album[30] = '\0';
	memcpy(mp3.id3.year, map + *fpos, 4);
	*fpos += 4;
	mp3.id3.year[4] = '\0';
	memcpy(mp3.id3.comment, map + *fpos, 30);
	mp3.id3.comment[30] = '\0';
	mp3.id3_isvalid = 1;
	unpad(mp3.id3.title);
	unpad(mp3.id3.artist);
	unpad(mp3.id3.album);
	unpad(mp3.id3.year);
	unpad(mp3.id3.comment);
}

static int find_valid_start(unsigned char *map, off_t numbytes, off_t *fpos)
{
	int frame_len;

	frame_len = get_header(map, numbytes, fpos, &mp3.header);
	if (frame_len < 0)
		return frame_len;
	if (!frame_len) {
		frame_len = mp3_seek_next_header(map, numbytes, fpos);
		if (frame_len <= 0)
			return frame_len;
	} else
		*fpos -= FRAME_HEADER_SIZE;
	if (frame_len <= 1)
		return -E_FRAME_LENGTH;
	return frame_len;
}

static int mp3_read_info(unsigned char *map, off_t numbytes,
		struct audio_format_info *afi)
{
	long fl_avg = 0, freq_avg = 0, br_avg = 0;
	int ret, len = 0, old_br = -1;
	struct timeval total_time = {0, 0};
	unsigned chunk_table_size = 1000; /* gets increased on demand */
	off_t fpos = 0;

	afi->chunks_total = 0;
	afi->chunk_table = para_malloc(chunk_table_size * sizeof(size_t));
	mp3_get_id3(map, numbytes, &fpos);
	fpos = 0;
	mp3.vbr = 0;
	while (1) {
		int freq, br, fl;
		struct timeval tmp, cct; /* current chunk time */
		fpos += len;
		len = find_valid_start(map, numbytes, &fpos);
		if (len <= 0)
			break;
		freq = header_frequency(&mp3.header);
		br = header_bitrate(&mp3.header);
		fl = frame_length(&mp3.header);
		if (freq < 0 || br < 0 || fl < 0)
			continue;
		tmp.tv_sec = fl;
		tmp.tv_usec = 0;
		tv_divide(br * 125, &tmp, &cct);
		tv_add(&cct, &total_time, &tmp);
		total_time = tmp;
		//PARA_DEBUG_LOG("%s: br: %d, freq: %d, fl: %d, cct: %lu\n", __func__, br, freq, fl, cct.tv_usec);
		if (afi->chunks_total >= chunk_table_size) {
			chunk_table_size *= 2;
			afi->chunk_table = para_realloc(afi->chunk_table,
				chunk_table_size * sizeof(size_t));
		}
		afi->chunk_table[afi->chunks_total] = fpos;
//		if (afi->chunks_total < 10 || !(afi->chunks_total % 1000))
//			PARA_INFO_LOG("chunk #%lu: %zd\n", afi->chunks_total,
//				afi->chunk_table[afi->chunks_total]);
		afi->chunks_total++;
		if (afi->chunks_total == 1) {
			freq_avg = freq;
			br_avg = br;
			old_br = br;
			fl_avg = fl;
			continue;
		}
		freq_avg += (freq - freq_avg) / (afi->chunks_total + 1);
		fl_avg += (fl - fl_avg) / (afi->chunks_total + 1);
		br_avg += (br - br_avg) / ((long)afi->chunks_total + 1);
		if (old_br != br)
			mp3.vbr = 1;
		old_br = br;
	}
	ret = -E_MP3_INFO;
	if (!afi->chunks_total || !freq_avg || !br_avg)
		goto err_out;
	afi->chunk_table[afi->chunks_total] = numbytes - 1;
	afi->bitrate = br_avg;
	afi->frequency = freq_avg;
	afi->channels = header_channels(&mp3.header);
	afi->seconds_total = (tv2ms(&total_time) + 500) / 1000;
	tv_divide(afi->chunks_total, &total_time, &afi->chunk_tv);
	PARA_DEBUG_LOG("%lu chunks, each %lums\n", afi->chunks_total,
		tv2ms(&afi->chunk_tv));
	tv_scale(3, &afi->chunk_tv, &afi->eof_tv);
	PARA_DEBUG_LOG("eof timeout: %lu\n", tv2ms(&afi->eof_tv));
	return 1;
err_out:
	PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
	free(afi->chunk_table);
	return ret;
}

/*
 * Read mp3 information from audio file
 */
static int mp3_get_file_info(char *map, off_t numbytes,
		struct audio_format_info *afi)
{
	int ret;

	ret = mp3_read_info((unsigned char *)map, numbytes, afi);
	if (ret < 0)
		return ret;
	write_info_str(afi);
	if (afi->seconds_total < 2 || !afi->chunks_total)
		return -E_MP3_INFO;
	return 1;
}

static const char* mp3_suffixes[] = {"mp3", NULL};

/**
 * the init function of the mp3 audio format handler
 *
 * \param afh pointer to the struct to initialize
 */
void mp3_init(struct audio_format_handler *afh)
{
	afh->get_file_info = mp3_get_file_info;
	afh->suffixes = mp3_suffixes;
}
